import unittest

from flask_monitoringdashboard.core.profiler.util import PathHash
from flask_monitoringdashboard.database import StackLine, CodeLine

FN = 'filename0'
LN = 42

STACK_LINES = [
            StackLine(request_id=0, position=0, indent=0, duration=1010,
                      code=CodeLine(filename=FN, line_number=0, function_name='None', code='f()')),
            StackLine(request_id=0, position=1, indent=1, duration=500,
                      code=CodeLine(filename=FN, line_number=1, function_name='f', code='sleep(1)')),
            StackLine(request_id=0, position=1, indent=1, duration=510,
                      code=CodeLine(filename=FN, line_number=2, function_name='f', code='sleep(1.01)'))]


class TestPathHash(unittest.TestCase):

    def setUp(self):
        self.path = PathHash()

    def test_get_path(self):
        self.assertEqual(self.path.get_path(FN, LN), '0:42')

    def test_append(self):
        self.path.get_path(FN, LN)
        self.assertEqual(self.path.append(FN, LN), '0:42->0:42')

    def test_encode(self):
        self.assertEqual(self.path._encode(FN, LN), '0:42')

    def test_decode(self):
        self.assertEqual(self.path._decode(self.path._encode(FN, LN)), (FN, LN))

    def test_get_indent(self):
        self.assertEqual(self.path.get_indent(''), 0)
        self.assertEqual(self.path.get_indent('0:42'), 1)
        self.assertEqual(self.path.get_indent('0:42->0:42'), 2)

    def test_get_last_fn_ln(self):
        self.assertEqual(self.path.get_last_fn_ln(self.path._encode(FN, LN)), (FN, LN))

    def test_get_code(self):
        for index, item in enumerate(STACK_LINES):
            path = self.path.get_stacklines_path(STACK_LINES, index)
            self.assertEqual(self.path.get_code(path), item.code.code)

    def test_get_stacklines_path(self):
        self.assertEqual(self.path.get_stacklines_path(STACK_LINES, 0), '1:0')
        self.assertEqual(self.path.get_stacklines_path(STACK_LINES, 1), '1:0->1:2')
        self.assertEqual(self.path.get_stacklines_path(STACK_LINES, 2), '1:0->1:3')



"""
    Contains all code for generating plots
    This files ensures consistency for all plots
"""
import plotly
import plotly.graph_objs as go
from flask_monitoringdashboard.core.plot.util import add_default_value
from flask_monitoringdashboard.core.plot.plots import heatmap, boxplot, barplot, scatter, get_average_bubble_size


def get_layout(**kwargs):
    """
    :param kwargs: additional arguments for the layout
    :return: a Plotly Layout object with the required values
    """
    kwargs = add_default_value('showlegend', False, **kwargs)
    kwargs = add_default_value('autosize', True, **kwargs)
    kwargs = add_default_value('plot_bgcolor', 'rgba(249,249,249,1)', **kwargs)
    kwargs = add_default_value('height', 700, **kwargs)
    kwargs = add_default_value('hovermode', 'closest', **kwargs)
    return go.Layout(**kwargs)


def get_margin(**kwargs):
    """
    :param kwargs: additional arguments for the Margin object
    :return: a Plotly Margin instance
    """
    kwargs = add_default_value('l', 200, **kwargs)
    return go.Margin(**kwargs)


def get_figure(layout, data, **kwargs):
    """
    :param layout: must be a Plotly Layout instance
    :param data: the data for the
    :param kwargs: additional arguments for the plot
    :return: A plotly generated plot with the required data
    """
    if not data:
        return None
    kwargs = add_default_value('output_type', 'div', **kwargs)
    kwargs = add_default_value('show_link', False, **kwargs)
    return plotly.offline.plot(go.Figure(data=data, layout=layout), **kwargs)



from flask_monitoringdashboard.core.profiler.util.pathHash import PathHash, LINE_SPLIT


def order_histogram(items, path=''):
    """
    Finds the order of self._text_dict and assigns this order to self._lines_body
    :param items: list of key, value. Obtained by histogram.items()
    :param path: used to filter the results
    :return The items, but sorted
    """
    sorted_list = []
    indent = PathHash.get_indent(path) + 1

    order = sorted([(key, value) for key, value in items
                    if key[0][:len(path)] == path and PathHash.get_indent(key[0]) == indent],
                   key=lambda row: row[0][1])
    for key, value in order:
        sorted_list.append((key, value))
        sorted_list.extend(order_histogram(items=items, path=key[0]))
    return sorted_list



"""
    This file can be executed for developing purposes.
    It is not used when the flask_monitoring_dashboard is attached to an existing flask application.
"""
import random

from flask import Flask


def create_app():
    import flask_monitoringdashboard as dashboard
    import time

    app = Flask(__name__)

    dashboard.config.outlier_detection_constant = 0
    # dashboard.config.database_name = 'sqlite:///flask_monitoring_dashboard_v10.db'
    dashboard.config.version = '3.1'
    dashboard.bind(app)

    def f():
        time.sleep(2)
        time.sleep(1)

    @app.route('/endpoint')
    def endpoint():
        f()
        return 'Ok'

    @app.route('/endpoint1')
    def endpoint1():
        f()
        return 'Ok'

    @app.route('/endpoint2')
    def endpoint2():
        f()
        return 'Ok'

    @app.route('/endpoint3')
    def endpoint3():
        f()
        return 'Ok'

    @app.route('/endpoint4')
    def endpoint4():
        f()
        return 'Ok'

    @app.route('/endpoint5')
    def endpoint5():
        f()
        return 'Ok'

    @app.route('/endpoint6')
    def endpoint6():
        f()
        return 'Ok'

    @app.route('/endpoint7')
    def endpoint7():
        f()
        return 'Ok'

    @app.route('/endpoint8')
    def endpoint8():
        f()
        return 'Ok'

    @app.route('/endpoint9')
    def endpoint9():
        f()
        return 'Ok'

    @app.route('/endpoint10')
    def endpoint10():
        f()
        return 'Ok'

    return app


if __name__ == '__main__':
    create_app().run(debug=True, threaded=True)



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.forms import get_slider_form
from flask_monitoringdashboard.core.plot import get_layout, get_figure, get_margin, heatmap
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.database import Request, session_scope
from flask_monitoringdashboard.database.count import count_versions
from flask_monitoringdashboard.database.count_group import count_requests_group, get_value
from flask_monitoringdashboard.database.endpoint import get_endpoints
from flask_monitoringdashboard.database.versions import get_versions

TITLE = 'Multi Version API Utilization'

AXES_INFO = '''The X-axis presents the versions that are used. The Y-axis presents the 
endpoints that are found in the Flask application.'''

CONTENT_INFO = '''The color of the cell presents the distribution of the amount of requests that the 
application received in a single version for a single endpoint. The darker the cell, the more requests 
a certain endpoint has processed in that version. Since it displays the distribution of the load, each 
column sums up to 100%. This information can be used to validate which endpoints processes the most 
requests.'''


@blueprint.route('/version_usage', methods=['GET', 'POST'])
@secure
def version_usage():
    with session_scope() as db_session:
        form = get_slider_form(count_versions(db_session), 'Select the number of versions')
        graph = version_usage_graph(db_session, form)
    return render_template('fmd_dashboard/graph.html', graph=graph, title=TITLE,
                           information=get_plot_info(AXES_INFO, CONTENT_INFO), form=form)


def version_usage_graph(db_session, form):
    """
    Used for getting a Heatmap with an overview of which endpoints are used in which versions
    :param db_session: session for the database
    :param form: instance of SliderForm
    :return:
    """
    endpoints = get_endpoints(db_session)
    versions = get_versions(db_session, limit=form.get_slider_value())

    requests = [count_requests_group(db_session, Request.version_requested == v) for v in versions]
    total_hits = []
    hits = [[]] * len(endpoints)

    for hits_version in requests:
        total_hits.append(max(1, sum([value for key, value in hits_version])))

    for j in range(len(endpoints)):
        hits[j] = [0] * len(versions)
        for i in range(len(versions)):
            hits[j][i] = get_value(requests[i], endpoints[j].id) * 100 / total_hits[i]

    layout = get_layout(
        xaxis={'title': 'Versions', 'type': 'category'},
        yaxis={'type': 'category', 'autorange': 'reversed'},
        margin=get_margin()
    )

    trace = heatmap(
        z=hits,
        x=versions,
        y=['{} '.format(e.name) for e in endpoints],
        colorbar={
            'titleside': 'top',
            'tickmode': 'array',
        }
    )
    return get_figure(layout=layout, data=[trace])



import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.data_grouped import get_endpoint_data_grouped, get_test_data_grouped, \
    get_version_data_grouped
from flask_monitoringdashboard.test.db.test_count_group import TestCountGroup
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, NAME, REQUESTS


class TestDataGrouped(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    @staticmethod
    def median(values):
        return sum(values) / len(values)

    def test_get_endpoint_data_grouped(self):
        with session_scope() as db_session:
            self.assertEqual(get_endpoint_data_grouped(db_session, self.median), {1: 12000}.items())

    def test_get_test_data_grouped(self):
        with session_scope() as db_session:
            TestCountGroup.add_test_data(db_session)
            self.assertEqual(get_test_data_grouped(db_session, self.median), {NAME: 1789.5}.items())

    def test_get_version_data_grouped(self):
        with session_scope() as db_session:
            self.assertEqual(get_version_data_grouped(db_session, lambda x: x), {'1.0': REQUESTS}.items())



"""
    Helper function for adding an argument to the **kwargs
"""


def add_default_value(arg_name, value, **kwargs):
    """ Add argument if it is not in the kwargs already """
    if arg_name not in kwargs.keys():
        kwargs[arg_name] = value
    return kwargs





"""
    Core files for the Flask Monitoring Dashboard. It contains the following packages:
    - config handles the configuration-file
    - forms handles generating and processing WTF-forms
    - plot handles the plotly library
    - profiler handles profiling requests

    It also contains the following files
    - auth.py handles authentication
    - colors.py handles color-hash
    - group_by.py handles the group_by functionality
    - info_box.py handles the generation of information box, one for each graph
    - measurements.py contains a number of wrappers, one for each monitoring level
    - rules.py includes a function that retrieves all rules of the entire Flask
      application, or for a specific endpoint.
    - timezone: handles utc-timezone <==> local-timezone
    - utils: for other functions
"""


import threading
import unittest

from flask_monitoringdashboard.core.profiler import StacktraceProfiler
from flask_monitoringdashboard.database import Endpoint
from flask_monitoringdashboard.test.utils import NAME, set_test_environment, clear_db, add_fake_data, get_test_app


class TestStacktraceProfiler(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()
        self.app = get_test_app()

    def test_after_run(self):
        with self.app.test_request_context():
            from flask import request
            request.environ['REMOTE_ADDR'] = '127.0.0.1'
            current_thread = threading.current_thread().ident
            ip = request.environ['REMOTE_ADDR']
            thread = StacktraceProfiler(current_thread, Endpoint(id=0, name=NAME), ip)
            thread._keeprunning = False
            self.assertRaises(ValueError, thread.run)



from flask import request
from flask_wtf import FlaskForm
from wtforms import SubmitField
from wtforms.fields.html5 import IntegerRangeField

DEFAULT_SLIDER_VALUE = 10


class SliderForm(FlaskForm):
    """
        Class for generating a slider that can be used to reduce the graph.
    """
    slider = IntegerRangeField()
    submit = SubmitField('Update')
    title = 'Select a number below for reducing the size of the graph'

    def get_slider_value(self):
        """
        :return: the value from the slider
        """
        if 'slider' in request.form:
            return request.form['slider']
        return self.start_value

    def content(self):
        return '''
          <div class="row">
            <div class="col-sm-3"><span>{}</span></div>
            <div class="col-sm-3" style="text-align: center"><output id="outputValue">{}</output></div>
            <div class="col-sm-3" style="text-align: right"><span>{}</span></div>
          </div>
          <div class="row">
            <div class="col-sm-9">{}</div>
            <div class="col-sm-3">{}</div>
          </div>'''.format(self.min_value, self.start_value, self.max_value,
                           self.slider(style="width: 100%;", oninput="outputValue.value=this.value",
                                       min=self.min_value, max=self.max_value, value=self.start_value),
                           self.submit(class_="btn btn-primary btn-block"))


def get_slider_form(slider_max=100, title=None):
    """
    Return a SliderForm with the range from 0 to slider_max
    :param slider_max: maximum value for the slider
    :param title: override the default title
    :return: a SliderForm with the range (0 ... slider_max)
    """
    form = SliderForm(request.form)
    form.min_value = 1
    form.max_value = slider_max
    if title:
        form.title = title
    if 'slider' in request.form:
        form.start_value = request.form['slider']
    else:
        form.start_value = min(DEFAULT_SLIDER_VALUE, slider_max)
    return form



import datetime

from flask import make_response, render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import admin_secure
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.request import get_data

CSV_COLUMNS = ['endpoint_id', 'duration', 'time_requested', 'version_requested', 'group_by', 'ip']


@blueprint.route('/download-csv')
@admin_secure
def download_csv():
    csv = ','.join(CSV_COLUMNS) + '\n'
    with session_scope() as db_session:
        for entry in get_data(db_session):
            csv += ','.join([str(entry.__getattribute__(c)) for c in CSV_COLUMNS]) + '\n'

    response = make_response(csv)
    response.headers["Content-Disposition"] = "attachment; filename=measurements_{0}.csv".format(
        str(datetime.datetime.utcnow()).replace(" ", "_").replace(":", "-")[:19])
    return response


@blueprint.route('/view-csv')
@admin_secure
def view_csv():
    csv = [','.join(CSV_COLUMNS)]
    with session_scope() as db_session:
        for entry in get_data(db_session):
            csv.append(','.join([str(entry.__getattribute__(c)) for c in CSV_COLUMNS]) + '\n')
    return render_template('fmd_export-data.html', data=csv)



from flask import request, render_template

from flask_monitoringdashboard import blueprint, user_app
from flask_monitoringdashboard.core.auth import admin_secure
from flask_monitoringdashboard.core.colors import get_color
from flask_monitoringdashboard.core.forms import get_monitor_form
from flask_monitoringdashboard.core.info_box import get_rules_info
from flask_monitoringdashboard.core.measurement import add_decorator
from flask_monitoringdashboard.core.rules import get_rules
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count_group import get_value
from flask_monitoringdashboard.database.endpoint import get_endpoint_by_name, update_endpoint, get_last_requested


@blueprint.route('/rules', methods=['GET', 'POST'])
@admin_secure
def rules():
    """
    Renders a table with all rules from the user_app. The fmd_dashboard rules are excluded
    In case of the POST request, the data from the form is validated and processed, such that the required rules are
    monitored
    :return:
    """
    if request.method == 'POST':
        with session_scope() as db_session:
            endpoint_name = request.form['name']
            value = int(request.form['value'])
            update_endpoint(db_session, endpoint_name, value=value)

            # Remove wrapper
            original = getattr(user_app.view_functions[endpoint_name], 'original', None)
            if original:
                user_app.view_functions[endpoint_name] = original

        with session_scope() as db_session:
            add_decorator(get_endpoint_by_name(db_session, endpoint_name))

        return 'OK'

    with session_scope() as db_session:
        last_accessed = get_last_requested(db_session)
        all_rules = []
        for rule in get_rules():
            db_rule = get_endpoint_by_name(db_session, rule.endpoint)
            all_rules.append({
                'color': get_color(rule.endpoint),
                'rule': rule.rule,
                'endpoint': rule.endpoint,
                'methods': rule.methods,
                'last_accessed': get_value(last_accessed, rule.endpoint, default=None),
                'form': get_monitor_form(rule.endpoint, db_rule.monitor_level)
            })
    return render_template('fmd_rules.html', rules=all_rules, information=get_rules_info())



from collections import defaultdict

from flask import render_template
import json

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.profiler.util.groupedStackLine import GroupedStackLine
from flask_monitoringdashboard.core.profiler.util.pathHash import PathHash
from flask_monitoringdashboard.core.utils import get_endpoint_details
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.stack_line import get_grouped_profiled_requests


@blueprint.route('/endpoint/<endpoint_id>/grouped-profiler')
@secure
def grouped_profiler(endpoint_id):
    with session_scope() as db_session:
        details = get_endpoint_details(db_session, endpoint_id)
        requests = get_grouped_profiled_requests(db_session, endpoint_id)
        db_session.expunge_all()
    total_duration = sum([r.duration for r in requests])
    histogram = defaultdict(list)  # path -> [list of values]
    path_hash = PathHash()

    for r in requests:
        for index, stack_line in enumerate(r.stack_lines):
            key = path_hash.get_stacklines_path(r.stack_lines, index)
            histogram[key].append(stack_line.duration)

    table = []
    for key, duration_list in sorted(histogram.items(), key=lambda row: row[0]):
        table.append(GroupedStackLine(indent=path_hash.get_indent(key) - 1, code=path_hash.get_code(key),
                                      values=duration_list, total_sum=total_duration, total_hits=len(requests)))
    for index, item in enumerate(table):
        table[index].compute_body(index, table)

    sunburst = json.dumps(table_to_json(table))
    return render_template('fmd_dashboard/profiler_grouped.html', details=details, table=table, sunburst=sunburst,
                           title='Grouped Profiler results for {}'.format(details['endpoint']))


def table_to_json(table, parent=None):
    if not parent:
        root_list = [row for row in table if row.indent == 0]
        if root_list:
            root = root_list[len(root_list)-1]
            return {
                'name': root.code,
                'children': table_to_json(table, parent=root)
            }
        else:
            return {}

    children = []
    for child in [table[index] for index in parent.body if table[index].indent == parent.indent + 1]:
        if child.body:
            children.append({
                'name': child.code,
                'children': table_to_json(table, child)
            })
        else:  # child is leaf
            children.append({
                'name': child.code,
                'size': child.sum
            })
    return children



import unittest

from flask_monitoringdashboard.core.profiler.util import order_histogram


class TestProfilerUtil(unittest.TestCase):

    def test_order_histogram(self):
        histogram = {
            ('0:42->1:12', 'c'): 610,
            ('0:42', 'a'): 1234,
            ('0:42->1:13', 'b'): 614
        }
        self.assertEqual(order_histogram(histogram.items()),
                         [(('0:42', 'a'), 1234), (('0:42->1:13', 'b'), 614), (('0:42->1:12', 'c'), 610)])



import unittest

from flask_monitoringdashboard.test.utils import get_test_app


class TestAuth(unittest.TestCase):

    def setUp(self):
        self.app = get_test_app()

    def test_check_login(self):
        """
            Test if function 'check_login' works
        """
        from flask_monitoringdashboard import config
        from flask_monitoringdashboard.core.auth import check_login

        with self.app.test_request_context():
            self.assertTrue(check_login(config.username, config.password))
            self.assertTrue(check_login(config.guest_username, config.guest_password[0]))



import datetime
import unittest

from flask_monitoringdashboard.core.timezone import to_local_datetime, to_utc_datetime


class TestTimezone(unittest.TestCase):

    def test_timezone(self):
        dt = datetime.datetime.now()
        self.assertEqual(to_local_datetime(to_utc_datetime(dt)), dt)
        self.assertEqual(to_utc_datetime(to_local_datetime(dt)), dt)

    def test_timezone_none(self):
        self.assertEqual(to_local_datetime(None), None)
        self.assertEqual(to_utc_datetime(None), None)



"""
Contains all functions that returns results of all tests
"""
from sqlalchemy import func

from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.database import Endpoint, Test, TestResult, TestEndpoint


def add_or_update_test(db_session, name, passing, last_tested, version_added):
    """ Add a unit test or update it. """
    test = db_session.query(Test).filter(Test.name == name).first()
    if test:
        test.name = name
        test.passing = passing
        test.last_tested = last_tested
    else:
        db_session.add(Test(name=name, passing=passing, last_tested=last_tested, version_added=version_added))
    db_session.commit()


def add_test_result(db_session, name, duration, time, version, job_id, iteration):
    """ Add a test result to the database. """
    test_id = db_session.query(Test).filter(Test.name == name).first().id
    db_session.add(TestResult(test_id=test_id, duration=duration, time_added=time, app_version=version,
                              travis_job_id=job_id, run_nr=iteration))


def get_sorted_job_ids(db_session, column, limit):
    """ Returns a decreasing sorted list of the Travis job ids. """
    query = db_session.query(column).group_by(column)
    if limit:
        query = query.limit(limit)
    return sorted([float(build[0]) for build in query.all()], reverse=True)


def get_test_suites(db_session, limit=None):
    """ Returns a decreasing sorted list of Travis job ids that collected Test data. """
    return get_sorted_job_ids(db_session, TestResult.travis_job_id, limit)


def get_travis_builds(db_session, limit=None):
    """ Returns a decreasing sorted list of Travis job ids that collected Endpoint data. """
    return get_sorted_job_ids(db_session, TestEndpoint.travis_job_id, limit)


def get_suite_measurements(db_session, suite):
    """ Return all measurements for some Travis build. Used for creating a box plot. """
    result = [result[0] for result in
              db_session.query(TestResult.duration).filter(TestResult.travis_job_id == str(suite)).all()]
    return result if len(result) > 0 else [0]


def get_endpoint_measurements(db_session, suite):
    """ Return all measurements for some Travis build. Used for creating a box plot. """
    result = [result[0] for result in
              db_session.query(TestEndpoint.duration).filter(TestEndpoint.travis_job_id == str(suite)).all()]
    return result if len(result) > 0 else [0]


def get_endpoint_measurements_job(db_session, name, job_id):
    """ Return all measurements for some test of some Travis build. Used for creating a box plot. """
    endpoint_id = db_session.query(Endpoint.id).filter(Endpoint.name == name).first()[0]
    result = db_session.query(TestEndpoint).join(TestEndpoint.endpoint).filter(
        TestEndpoint.endpoint_id == endpoint_id).filter(TestEndpoint.travis_job_id == job_id).all()
    return [r.duration for r in result] if len(result) > 0 else [0]


def get_last_tested_times(db_session):
    """ Returns the last tested time of each of the endpoints. """
    results = db_session.query(TestEndpoint, func.max(TestEndpoint.time_added)).join(
        TestEndpoint.endpoint).group_by(TestEndpoint.endpoint_id).all()
    return [(result[0].endpoint.name, to_local_datetime(result[1])) for result in results]



from flask_monitoringdashboard.core.profiler.util.stringHash import StringHash

STRING_SPLIT = '->'
LINE_SPLIT = ':'


class PathHash(object):
    """
    Used for encoding the stacktrace.
    A stacktrace can be seen by a list of tuples (filename and linenumber): e.g. [(fn1, 25), (fn2, 30)]
    this is encoded as a string:

        encoded = 'fn1:25->fn2->30'

   However, the filename could possibly contain '->', therefore the filename is hashed into a number.
    So, the encoding becomes:

        encoded = '0:25->1:30'
    """

    def __init__(self):
        self._string_hash = StringHash()
        self._current_path = ''

    def set_path(self, path):
        self._current_path = path

    def get_path(self, fn, ln):
        """
        :param fn: String with the filename
        :param ln: line number
        :return: Encoded path name.
        """
        self._current_path = self.append(fn, ln)
        return self._current_path

    def append(self, fn, ln):
        """
        Concatenate the current_path with the new path.
        :param fn: filename
        :param ln: line number
        :return: The new current_path
        """
        if self._current_path:
            return self._current_path + STRING_SPLIT + self._encode(fn, ln)
        return self._encode(fn, ln)

    def _encode(self, fn, ln):
        """
        Encoded fn and ln in the following way:
            _encode(fn, ln) => hash(fn):ln
        :param fn: filename (string)
        :param ln: linenumber (int)
        :return: String with the hashed filename, and linenumber
        """

        return str(self._string_hash.hash(fn)) + LINE_SPLIT + str(ln)

    def _decode(self, string):
        """ Opposite of _encode

        Example: _decode('0:12') => ('fn1', 12)
        """
        hash, ln = string.split(LINE_SPLIT)
        return self._string_hash.unhash(int(hash)), int(ln)

    @staticmethod
    def get_indent(path):
        """
        Compute the amount of callers given a path.
        :param path: the path from the root to the statement
        :return: an integer
        """
        if path:
            return len(path.split(STRING_SPLIT))
        return 0

    def get_code(self, path):
        """
        :param path: only take the last tuple of the path. the last part contains the code line, but hashed.
        :return: the line of code, based on the given path
        """
        last = path.rpartition(STRING_SPLIT)[-1]
        return self._string_hash.unhash(int(last.split(LINE_SPLIT, 1)[1]))

    def get_last_fn_ln(self, string):
        last = string.rpartition(STRING_SPLIT)[-1]
        return self._decode(last)

    def get_stacklines_path(self, stack_lines, index):
        """
        Used for grouping multiple requests
        :param stack_lines: list of StackLine objects.
        :param index: index in the stack_lines, so 0 <= index < len(stack_lines)
        :return: the StackLinePath that belongs to the given index
        """
        self.set_path('')
        path = []
        while index >= 0:
            path.append(stack_lines[index].code)
            current_indent = stack_lines[index].indent
            while index >= 0 and stack_lines[index].indent != current_indent - 1:
                index -= 1
        for code_line in reversed(path):
            self._current_path = self.append(code_line.filename, self._string_hash.hash(code_line.code))
        return self._current_path



import configparser
import os

import pytz

from flask_monitoringdashboard.core.config.parser import parse_string, parse_version, parse_bool, parse_literal
from tzlocal import get_localzone


class Config(object):
    """
        The settings can be changed by setting up a config file. For an example of a config file, see
         config.cfg in the main-directory. 
    """

    def __init__(self):
        """
            Sets the default values for the project
        """
        # dashboard
        self.version = '1.0'
        self.link = 'dashboard'
        self.monitor_level = 3
        self.outlier_detection_constant = 2.5
        self.sampling_period = 0

        # database
        self.database_name = 'sqlite:///flask_monitoringdashboard.db'
        self.table_prefix = ''

        # authentication
        self.username = 'admin'
        self.password = 'admin'
        self.guest_username = 'guest'
        self.guest_password = ['guest_password']
        self.security_token = 'cc83733cb0af8b884ff6577086b87909'

        # visualization
        self.colors = {}
        self.timezone = pytz.timezone(str(get_localzone()))

        # define a custom function to retrieve the session_id or username
        self.group_by = None

    def init_from(self, file=None, envvar=None):
        """
            The config_file must at least contains the following variables in section 'dashboard':
            - APP_VERSION: the version of the app that you use. Updating the version helps in
                showing differences in execution times of a function over a period of time.
            - GIT = If you're using git, then it is easier to set the location to the .git-folder,
                The location is relative to the config-file.
            - CUSTOM_LINK: The dashboard can be visited at localhost:5000/{CUSTOM_LINK}.
            - MONITOR_LEVEL: The level for monitoring your endpoints. The default value is 3.
            - OUTLIER_DETECTION_CONSTANT: When the execution time is more than this constant *
                average, extra information is logged into the database. A default value for this
                variable is 2.5.
            - SAMPLING_PERIOD: Time between two profiler-samples. The time must be specified in ms.
                If this value is not set, the profiler continuously monitors.

            The config_file must at least contains the following variables in section 'authentication':
            - USERNAME: for logging into the dashboard, a username and password is required. The
                username can be set using this variable.
            - PASSWORD: same as for the username, but this is the password variable.
            - GUEST_USERNAME: A guest can only see the results, but cannot configure/download data.
            - GUEST_PASSWORD: A guest can only see the results, but cannot configure/download data.
            - SECURITY_TOKEN: Used for getting the data in /get_json_data

            The config_file must at least contains the following variables in section 'database':
            - DATABASE: Suppose you have multiple projects where you're working on and want to
                separate the results. Then you can specify different database_names, such that the
                result of each project is stored in its own database.
            - TABLE_PREFIX: A prefix to every table that the Flask-MonitoringDashboard uses, to ensure
                that there are no conflicts with the user of the dashboard.

            The config_file must at least contains the following variables in section 'visualization':
            - TIMEZONE: The timezone for converting a UTC timestamp to a local timestamp.
                for a list of all timezones, use the following:

                >>> import pytz  # pip install pytz
                >>> print(pytz.all_timezones)

            - COLORS: A dictionary to override the colors used per endpoint.

            :param file: a string pointing to the location of the config-file.
            :param envvar: a string specifying which environment variable holds the config file location
        """

        if envvar:
            file = os.getenv(envvar)
        if not file:
            # Travis does not need a config file.
            if '/home/travis/build/' in os.getcwd():
                return
            print("No configuration file specified. Please do so.")
            return

        parser = configparser.RawConfigParser()
        try:
            parser.read(file)

            # parse 'dashboard'
            self.version = parse_version(parser, 'dashboard', self.version)
            self.link = parse_string(parser, 'dashboard', 'CUSTOM_LINK', self.link)
            self.monitor_level = parse_literal(parser, 'dashboard', 'MONITOR_LEVEL', self.monitor_level)
            self.outlier_detection_constant = parse_literal(parser, 'dashboard', 'OUTlIER_DETECTION_CONSTANT',
                                                            self.outlier_detection_constant)
            self.sampling_period = parse_literal(parser, 'dashboard', 'SAMPLING_RATE', self.sampling_period) / 1000

            # parse 'authentication'
            self.username = parse_string(parser, 'authentication', 'USERNAME', self.username)
            self.password = parse_string(parser, 'authentication', 'PASSWORD', self.password)
            self.security_token = parse_string(parser, 'authentication', 'SECURITY_TOKEN', self.security_token)
            self.guest_username = parse_string(parser, 'authentication', 'GUEST_USERNAME', self.guest_username)
            self.guest_password = parse_literal(parser, 'authentication', 'GUEST_PASSWORD', self.guest_password)

            # database
            self.database_name = parse_string(parser, 'database', 'DATABASE', self.database_name)
            self.table_prefix = parse_string(parser, 'database', 'TABLE_PREFIX', self.table_prefix)

            # visualization
            self.colors = parse_literal(parser, 'visualization', 'COLORS', self.colors)
            self.timezone = pytz.timezone(parse_string(parser, 'visualization', 'TIMEZONE', self.timezone.zone))
        except configparser.Error:
            raise



import sys
import threading
import time
import traceback

import psutil
from flask import request

from flask_monitoringdashboard import config
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.outlier import add_outlier
from flask_monitoringdashboard.database.request import get_avg_duration


class OutlierProfiler(threading.Thread):
    """
    Used for collecting additional information if the request is an outlier
    """

    def __init__(self, current_thread, endpoint):
        threading.Thread.__init__(self)
        self._current_thread = current_thread
        self._endpoint = endpoint
        self._stopped = False
        self._cpu_percent = ''
        self._memory = ''
        self._stacktrace = ''

        self._request = str(request.headers), str(request.environ), str(request.url)

    def run(self):
        # sleep for average * ODC ms
        with session_scope() as db_session:
            average = get_avg_duration(db_session, self._endpoint.id) * config.outlier_detection_constant
        time.sleep(average / 1000)
        if not self._stopped:
            stack_list = []
            try:
                frame = sys._current_frames()[self._current_thread]
            except KeyError:
                return
            in_endpoint_code = False
            # filename, line number, function name, source code line
            for fn, ln, fun, line in traceback.extract_stack(frame):
                if self._endpoint.name == fun:
                    in_endpoint_code = True
                if in_endpoint_code:
                    stack_list.append('File: "{}", line {}, in "{}": "{}"'.format(fn, ln, fun, line))

            # Set the values in the object
            self._stacktrace = '<br />'.join(stack_list)
            self._cpu_percent = str(psutil.cpu_percent(interval=None, percpu=True))
            self._memory = str(psutil.virtual_memory())

    def stop(self):
        self._stopped = True

    def add_outlier(self, request_id):
        if self._memory:
            with session_scope() as db_session:
                add_outlier(db_session, request_id, self._cpu_percent, self._memory, self._stacktrace, self._request)



"""
    Contains all views that present information about specific endpoints.
    The endpoints are the following:
    - heatmap/<endpoint_id>: shows a heatmap of the usage for <endpoint_id>
    - Time per version per user/<endpoint_id>: shows an overview of the time per user per version for a given
            <endpoint_id>
    - Time per version per ip/<endpoint_id>: shows an overview of the time per user per ip for a given <endpoint_id>
    - Time per version/<endpoint_id>: shows an overview of all requests per version for a given <endpoint_id>
    - Time per user/<endpoint_id>: shows an overview of all requests per user for a given <endpoint_id>
    - Outliers/<endpoint_id>: shows information about requests that take too long.
"""

from flask_monitoringdashboard.views.details.heatmap import endpoint_hourly_load
from flask_monitoringdashboard.views.details.outliers import outliers
from flask_monitoringdashboard.views.details.time_user import users
from flask_monitoringdashboard.views.details.time_version import versions
from flask_monitoringdashboard.views.details.version_ip import version_ip
from flask_monitoringdashboard.views.details.version_user import version_user
from flask_monitoringdashboard.views.details.profiler import profiler
from flask_monitoringdashboard.views.details.grouped_profiler import grouped_profiler



"""
Class used for hashing the paths.
"""


class StringHash(object):

    def __init__(self):
        self._h = {}

    def hash(self, string):
        """
        Performs the following reduction:

        hash('abc') ==> 0
        hash('def') ==> 1
        hash('abc') ==> 0

        :param string: the string to be hashed
        :return: a unique int for every string.
        """
        if string in self._h:
            return self._h[string]
        self._h[string] = len(self._h)
        return self._h[string]

    def unhash(self, hash):
        """ Opposite of hash.

        unhash(hash('abc')) == 'abc

        :param hash: string to be unhashed
        :return: the value that corresponds to the given hash
        """

        for k, v in self._h.items():
            if v == hash:
                return k

        raise ValueError('Value not possible to unhash: {}'.format(hash))



from flask import render_template

from flask_monitoringdashboard import blueprint
from flask_monitoringdashboard.core.auth import secure
from flask_monitoringdashboard.core.plot import get_layout, get_figure, boxplot, get_margin
from flask_monitoringdashboard.core.info_box import get_plot_info
from flask_monitoringdashboard.core.utils import simplify
from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.count_group import get_value
from flask_monitoringdashboard.database.data_grouped import get_endpoint_data_grouped
from flask_monitoringdashboard.database.endpoint import get_endpoints

TITLE = 'API Performance'

AXES_INFO = '''The X-axis presents the execution time in ms. The Y-axis presents every
endpoint of the Flask application.'''

CONTENT_INFO = '''In this graph, it is easy to compare the execution time of the different endpoints
across each other. This information can be used to validate which endpoints needs to be improved.'''


@blueprint.route('/endpoints')
@secure
def endpoints():
    return render_template('fmd_dashboard/graph.html', graph=endpoint_graph(), title=TITLE,
                           information=get_plot_info(AXES_INFO, CONTENT_INFO))


def endpoint_graph():
    """
    Creates a graph with the execution times per endpoint
    :return:
    """
    with session_scope() as db_session:
        data = get_endpoint_data_grouped(db_session, lambda x: simplify(x, 10))
        values = [boxplot(get_value(data, end.id, default=[]), name=end.name)
                  for end in get_endpoints(db_session)]

    layout = get_layout(
        height=350 + 40 * len(values),
        xaxis={'title': 'Execution time (ms)'},
        margin=get_margin()
    )
    return get_figure(layout, values)



"""
    Use this file for migrating the Database from v1.X.X to v2.X.X
    Before running the script, make sure to change the OLD_DB_URL and NEW_DB_URL on lines 9 and 10.
    Refer to http://docs.sqlalchemy.org/en/latest/core/engines.html on how to configure this.
"""
import datetime
from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, joinedload

from flask_monitoringdashboard.database import Endpoint, Request, Outlier, Test, TestResult

# OLD_DB_URL = 'dialect+driver://username:password@host:port/old_db'
# NEW_DB_URL = 'dialect+driver://username:password@host:port/new_db'
OLD_DB_URL = 'sqlite:////home/bogdan/databases/flask-dashboard.db'
NEW_DB_URL = 'sqlite:////home/bogdan/databases/flask-dashboard-2.0.db'
# OLD_DB_URL = 'mysql+pymysql://root:admin@localhost/migration1'
# NEW_DB_URL = 'mysql+pymysql://root:admin@localhost/migration2'


TABLES = ["rules", "functionCalls", "outliers", "tests", "testRun"]
DATE_FORMAT = "%Y-%m-%d %H:%M:%S.%f"
SEARCH_REQUEST_TIME = datetime.timedelta(seconds=10)

endpoint_dict = {}
outlier_dict = {}
tests_dict = {}


def create_new_db(db_url):
    from flask_monitoringdashboard.database import Base
    engine = create_engine(db_url)
    Base.metadata.drop_all(engine)
    Base.metadata.create_all(engine)
    Base.metadata.bind = engine
    global DBSession
    DBSession = sessionmaker(bind=engine)


def get_connection(db_url):
    engine = create_engine(db_url)
    connection = engine.connect()
    return connection


@contextmanager
def session_scope():
    """
    When accessing the database, use the following syntax:
        with session_scope() as db_session:
            db_session.query(...)

    :return: the session for accessing the database
    """
    session = DBSession()
    try:
        yield session
        session.commit()
    except Exception:
        session.rollback()
        raise
    finally:
        session.close()


def get_session(db_url):
    """This creates the new database and returns the session scope."""
    from flask_monitoringdashboard import config
    config.database_name = db_url

    import flask_monitoringdashboard.database
    return flask_monitoringdashboard.database.session_scope()


def parse(date_string):
    if not date_string:
        return None
    if isinstance(date_string, str):
        return datetime.datetime.strptime(date_string, DATE_FORMAT)
    return date_string


def move_rules(old_connection):
    rules = old_connection.execute("select * from {}".format(TABLES[0]))
    endpoints = []
    with session_scope() as db_session:
        for rule in rules:
            end = Endpoint(name=rule['endpoint'], monitor_level=rule['monitor'],
                           time_added=parse(rule['time_added']), version_added=rule['version_added'],
                           last_requested=parse(rule['last_accessed']))
            endpoints.append(end)
        db_session.bulk_save_objects(endpoints)


def populate_endpoint_dict(db_session):
    global endpoint_dict
    endpoints = db_session.query(Endpoint).all()
    for endpoint in endpoints:
        endpoint_dict[endpoint.name] = endpoint.id


def move_function_calls(old_connection):
    function_calls = old_connection.execute("select * from {}".format(TABLES[1]))
    requests = []
    with session_scope() as db_session:
        populate_endpoint_dict(db_session)
        for fc in function_calls:
            request = Request(endpoint_id=endpoint_dict[fc['endpoint']], duration=fc['execution_time'],
                              time_requested=parse(fc['time']), version_requested=fc['version'],
                              group_by=fc['group_by'], ip=fc['ip'])
            requests.append(request)
        db_session.bulk_save_objects(requests)


def get_request_id(requests, time, execution_time, start_index):
    for index, r in enumerate(requests):
        if index >= start_index:
            if abs(r.time_requested - parse(time)) < SEARCH_REQUEST_TIME and r.duration == execution_time:
                return r.id, index
    return None, start_index


def populate_outlier_dict(connection, db_session):
    global outlier_dict
    outliers = connection.execute("select * from {}".format(TABLES[2]))
    requests = db_session.query(Request).options(joinedload(Request.endpoint)).all()
    index = 0
    for outlier in outliers:
        req_id, index = get_request_id(requests, outlier['time'], outlier['execution_time'], start_index=index)
        outlier_dict[outlier['id']] = req_id


def move_outliers(old_connection):
    global outlier_dict
    old_outliers = old_connection.execute("select * from {}".format(TABLES[2]))
    outliers = []
    with session_scope() as db_session:
        populate_outlier_dict(old_connection, db_session)
        for o in old_outliers:
            outlier = Outlier(request_id=outlier_dict[o['id']], request_header=o['request_headers'],
                              request_environment=o['request_environment'], request_url=o['request_url'],
                              cpu_percent=o['cpu_percent'], memory=o['memory'], stacktrace=o['stacktrace'])
            outliers.append(outlier)
        db_session.bulk_save_objects(outliers)


def move_tests(old_connection):
    try:
        old_tests = old_connection.execute("select * from {}".format(TABLES[3]))
        tests = []
        with session_scope() as db_session:
            for t in old_tests:
                test = Test(name=t['name'], passing=t['succeeded'],
                            version_added='', last_tested=parse(t['lastRun']))
                tests.append(test)
            db_session.bulk_save_objects(tests)
    except Exception as err:
        print("tests table was not moved. Does the table exist?")
        print(err)


def populate_tests_dict(db_session):
    global tests_dict
    tests = db_session.query(Test).all()
    for test in tests:
        tests_dict[test.name] = test.id


def move_test_runs(old_connection):
    try:
        test_runs = old_connection.execute("select * from {}".format(TABLES[4]))
        test_results = []
        with session_scope() as db_session:
            populate_tests_dict(db_session)
            for tr in test_runs:
                test_result = TestResult(test_id=tests_dict[tr['name']], duration=tr['execution_time'],
                                         time_added=parse(tr['time']), app_version=tr['version'],
                                         travis_job_id=tr['suite'], run_nr=tr['run'])
                test_results.append(test_result)
            db_session.bulk_save_objects(test_results)
    except Exception as err:
        print("testRun table was not moved.")
        print(err)


def main():
    create_new_db(NEW_DB_URL)
    old_connection = get_connection(OLD_DB_URL)
    import timeit
    start = timeit.default_timer()
    move_rules(old_connection)
    t1 = timeit.default_timer()
    print("Moving rules took %f seconds" % (t1 - start))
    move_function_calls(old_connection)
    t2 = timeit.default_timer()
    print("Moving functionCalls took %f seconds" % (t2 - t1))
    move_outliers(old_connection)
    t3 = timeit.default_timer()
    print("Moving outliers took %f seconds" % (t3 - t2))
    move_tests(old_connection)
    t4 = timeit.default_timer()
    print("Moving tests took %f seconds" % (t4 - t3))
    move_test_runs(old_connection)
    t5 = timeit.default_timer()
    print("Moving testRuns took %f seconds" % (t5 - t4))

    print("Total time was %f seconds" % (t5 - start))


if __name__ == "__main__":
    main()



"""
Contains all functions that access an Endpoint object
"""
import datetime
from collections import defaultdict

from sqlalchemy import func, desc
from sqlalchemy.orm.exc import NoResultFound

from flask_monitoringdashboard.core.timezone import to_local_datetime
from flask_monitoringdashboard.database import Request, Endpoint


def get_num_requests(db_session, endpoint_id, start_date, end_date):
    """
    Returns a list with all dates on which an endpoint is accessed.
    :param db_session: session for the database
    :param endpoint_id: if None, the result is the sum of all endpoints
    :param start_date: datetime.date object
    :param end_date: datetime.date object
    :return list of dates
    """
    query = db_session.query(Request.time_requested)
    if endpoint_id:
        query = query.filter(Request.endpoint_id == endpoint_id)
    result = query.filter(Request.time_requested >= start_date, Request.duration <= end_date).all()

    return group_request_times([r[0] for r in result])


def group_request_times(datetimes):
    """
    Returns a list of tuples containing the number of hits per hour
    :param datetimes: list of datetime objects
    :return list of tuples ('%Y-%m-%d %H:00:00', count)
    """
    hours_dict = defaultdict(int)
    for dt in datetimes:
        round_time = dt.strftime('%Y-%m-%d %H:00:00')
        hours_dict[round_time] += 1
    return hours_dict.items()


def get_users(db_session, endpoint_id, limit=None):
    """
    Returns a list with the distinct group-by from a specific endpoint. The limit is used to filter the most used
    distinct.
    :param db_session: session for the database
    :param endpoint_id: the id of the endpoint to filter on
    :param limit: the max number of results
    :return a list with the group_by as strings.
    """
    query = db_session.query(Request.group_by, func.count(Request.group_by)). \
        filter(Request.endpoint_id == endpoint_id).group_by(Request.group_by). \
        order_by(desc(func.count(Request.group_by)))
    if limit:
        query = query.limit(limit)
    result = query.all()
    db_session.expunge_all()
    return [r[0] for r in result]


def get_ips(db_session, endpoint_id, limit=None):
    """
    Returns a list with the distinct group-by from a specific endpoint. The limit is used to filter the most used
    distinct.
    :param db_session: session for the database
    :param endpoint_id: the endpoint_id to filter on
    :param limit: the number of
    :return a list with the group_by as strings.
    """
    query = db_session.query(Request.ip, func.count(Request.ip)). \
        filter(Request.endpoint_id == endpoint_id).group_by(Request.ip). \
        order_by(desc(func.count(Request.ip)))
    if limit:
        query = query.limit(limit)
    result = query.all()
    db_session.expunge_all()
    return [r[0] for r in result]


def get_endpoint_by_name(db_session, endpoint_name):
    """
    Returns the Endpoint object from a given endpoint_name.
    If the result doesn't exist in the database, a new row is added.
    :param db_session: session for the database
    :param endpoint_name: string with the endpoint name
    :return Endpoint object
    """
    try:
        result = db_session.query(Endpoint). \
            filter(Endpoint.name == endpoint_name).one()
        result.time_added = to_local_datetime(result.time_added)
        result.last_requested = to_local_datetime(result.last_requested)
    except NoResultFound:
        result = Endpoint(name=endpoint_name)
        db_session.add(result)
        db_session.flush()
    db_session.expunge(result)
    return result


def get_endpoint_by_id(db_session, endpoint_id):
    """
    Returns the Endpoint object from a given endpoint id.
    :param db_session: session for the database
    :param endpoint_id: id of the endpoint.
    :return Endpoint object
    """
    result = db_session.query(Endpoint).filter(Endpoint.id == endpoint_id).one()
    db_session.expunge(result)
    return result


def update_endpoint(db_session, endpoint_name, value):
    """
    Updates the value of a specific Endpoint.
    :param db_session: session for the database
    :param endpoint_name: name of the endpoint
    :param value: new monitor level
    """
    db_session.query(Endpoint).filter(Endpoint.name == endpoint_name). \
        update({Endpoint.monitor_level: value})
    db_session.flush()


def get_last_requested(db_session):
    """
    Returns the accessed time of all endpoints.
    :param db_session: session for the database
    :return list of tuples with name of the endpoint and date it was last used
    """
    result = db_session.query(Endpoint.name, Endpoint.last_requested).all()
    db_session.expunge_all()
    return [(end, to_local_datetime(time)) for end, time in result]


def update_last_accessed(db_session, endpoint_name):
    """
    Updates the timestamp of last access of the endpoint.
    :param db_session: session for the database
    :param endpoint_name: name of the endpoint
    """
    db_session.query(Endpoint).filter(Endpoint.name == endpoint_name). \
        update({Endpoint.last_requested: datetime.datetime.utcnow()})


def get_endpoints(db_session):
    """
    Returns all Endpoint objects from the database.
    :param db_session: session for the database
    :return list of Endpoint objects
    """
    return db_session.query(Endpoint).all()



import unittest
from math import sqrt


class TestGroupedStackLine(unittest.TestCase):

    def test_grouped_stackline(self):
        from flask_monitoringdashboard.core.profiler.util.groupedStackLine import GroupedStackLine
        grouped_stackline = GroupedStackLine(indent=0, code='code', values=[10, 10, 40], total_sum=100, total_hits=6)
        self.assertEqual(grouped_stackline.hits, 3)
        self.assertEqual(grouped_stackline.sum, 60)
        self.assertEqual(grouped_stackline.standard_deviation, sqrt(200))
        self.assertEqual(grouped_stackline.hits_percentage, .5)
        self.assertEqual(grouped_stackline.percentage, .6)
        self.assertEqual(grouped_stackline.average, 20)


import unittest

from flask_monitoringdashboard.database import session_scope
from flask_monitoringdashboard.database.tested_endpoints import get_tested_endpoint_names
from flask_monitoringdashboard.test.db.test_count_group import TestCountGroup
from flask_monitoringdashboard.test.utils import set_test_environment, clear_db, add_fake_data, NAME


class TestTestedEndpoints(unittest.TestCase):

    def setUp(self):
        set_test_environment()
        clear_db()
        add_fake_data()

    def test_get_tested_endpoint_names(self):
        with session_scope() as db_session:
            TestCountGroup.add_test_data(db_session)
            self.assertEqual(get_tested_endpoint_names(db_session), [NAME])



from flask_monitoringdashboard import config

PRIMITIVES = (bool, bytes, float, int, str)


def recursive_group_by(argument):
    """
    Returns the result of the given argument. The result is computed as:
    - If the argument is a primitive (i.e. str, bool, int, ...) return its value.
    - If the argument is a function, call the function.
    - If the argument is iterable (i.e. list or tuple), compute the result by iterating over the argument
    Return type is always a string
    """

    if type(argument) in PRIMITIVES:
        return str(argument)

    if callable(argument):
        return recursive_group_by(argument())

    # Try if the argument is iterable (i.e. tuple or list)
    try:
        result_list = [recursive_group_by(i) for i in argument]
        result_string = ','.join(result_list)
        return '({})'.format(result_string)
    except TypeError:
        # Cannot deal with this
        return str(argument)


def get_group_by():
    """
    :return: a string with the value
    """
    group_by = None
    try:
        if config.group_by:
            group_by = recursive_group_by(config.group_by)
    except Exception as e:
        print('Can\'t execute group_by function: {}'.format(e))
    return str(group_by)